from unittest.mock import patch

from django.contrib.auth import get_user_model
from rest_framework.test import APIClient, APIRequestFactory

from .. import custom_exceptions
from ..authentication import LearngualAuthentication, LearngualFakeAuthentication
from .request_mock import requests as mock_requests

factory = APIRequestFactory()
User = get_user_model()


@patch("iam_service.learngual.authentication.get_user")
@patch("requests.get", return_value=mock_requests.get)
def test_authentication(mock_req, get_user, user: User, client: APIClient):  # type: ignore
    # Create a sample request
    get_user.return_value = user
    mock_req.json.return_value = {
        "account": {
            "id": "84bcaf2972",
            "cover_photo": None,
            "profile_photo": None,
            "type": "PERSONNAL",
            "metadata": {},
            "created_at": "2023-01-13T16:33:52.084540Z",
            "updated_at": "2023-01-13T16:33:52.084576Z",
        },
        "email": "Bulah53@gmail.com",
        "first_name": "Caitlyn",
        "id": "40e0e7013f",
        "last_name": "Marquardt",
        "registration_step": "REGISTRATION_COMPLETED",
        "username": "Eloisa.Senger42",
    }
    request = factory.get("/api/my-endpoint/")

    # Add the custom header to the request
    token = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
    token += (
        ".eyJ1c2VybmFtZSI6InRlc3R1c2VyIiwiaWF0IjoxNjIzMzQ5NTMwLCJleHAiOjE2MjMzNTM5MzB9"
    )
    token += ".6FcybZUxW1szsiavTbF-MfnEvC57lU3J0C1dd6nM2O0"
    request.META["HTTP_AUTHORIZATION"] = "Bearer " + token

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualAuthentication()

    # Authenticate the request using the custom authentication class
    authenticated = auth.authenticate(request)

    # Assert that authentication is successful
    assert authenticated[0] == user
    assert authenticated[1] == token


def test_fake_authentication_account(personal_account: object, client: APIClient):  # type: ignore

    request = factory.get("/api/my-endpoint/")

    request.META["HTTP_ACCOUNT"] = personal_account.id

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualFakeAuthentication()

    # Authenticate the request using the custom authentication class
    authenticated = auth.authenticate(request)
    assert authenticated, "authentication return None"

    # Assert that authentication is successful
    assert authenticated[0] == getattr(personal_account, "owner", None) or getattr(
        personal_account, "user", None
    )
    assert authenticated[0].account == personal_account
    assert authenticated[1] in ["fake_token", ""]


def test_fake_authentication_user(user: User):  # type: ignore

    request = factory.get("/api/my-endpoint/")

    request.META["HTTP_ACCOUNT"] = user.id

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualFakeAuthentication()

    # Authenticate the request using the custom authentication class
    authenticated = auth.authenticate(request)
    assert authenticated, "authentication return None"

    # Assert that authentication is successful
    assert authenticated[0] == user
    assert authenticated[0].account is None
    assert authenticated[1] in ["fake_token", ""]


def test_authentication_service_key(settings):

    request = factory.get("/api/my-endpoint/")

    # Add the custom header to the request

    request.META["HTTP_SERVICE_KEY"] = settings.LEARNGUAL_SERVICE_API_KEY

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualAuthentication()

    # Authenticate the request using the custom authentication class
    authenticated = auth.authenticate(request)

    # Assert that authentication is successful
    assert getattr(authenticated[0], "is_service", None)
    assert authenticated[1] == settings.LEARNGUAL_SERVICE_API_KEY


def test_authentication_invalid_service_key(settings):

    request = factory.get("/api/my-endpoint/")

    # Add the custom header to the request

    request.META["HTTP_SERVICE_KEY"] = "hshsggfaas"

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualAuthentication()

    try:
        # Authenticate the request using the custom authentication class
        authenticated = auth.authenticate(request)
        assert authenticated
    except custom_exceptions.UnAuthenticated:
        assert True



import json
import re
from pathlib import Path
from typing import Any

from django.conf import settings

from .translator import Translator


class TransformTemplateContext:
    def __init__(
        self,
        template_dir: str | Path,
        user_or_account: Any,
        kwargs_map: dict[str, dict[str, Any]] = {},
    ):
        """
        Initialize the TransformContext.

        :param json_file_path: Path to the JSON file.
        :param translate: Function to translate text.
        :param user_or_account: User or account object containing language preferences.
        :param kwargs_map: Dictionary containing key-value pairs for formatting.
        """
        self.kwargs_map = kwargs_map
        self.translate = Translator(
            target_language=user_or_account.language
        ).get_translation
        self.user_or_account = user_or_account
        self.template_dir = template_dir.strip("/")
        self.template_root = Path(settings.ROOT_DIR) / "templates"
        self.data = self.read_json(self.template_info_path)

    @property
    def template_file_path(self) -> Path:
        return self.template_root / self.template_dir / "index.html"

    @property
    def template(self) -> str:
        """template name

        Returns:
            str: _description_
        """
        return self.template_dir + "/index.html"

    @property
    def template_info_path(self) -> Path:
        return self.template_root / self.template_dir / "info.json"

    def read_json(self, file_path: str) -> dict[str, Any]:
        """
        Read JSON data from a file.

        :param file_path: Path to the JSON file.
        :return: Parsed JSON data as a dictionary.
        """
        with open(file_path) as file:
            return json.load(file)

    def extract_patterns(self, input_string: str) -> dict[str, str]:
        """
        Extract key-value patterns from the input string.

        :param input_string: String containing the patterns in the format {key: value}.
        :return: Dictionary containing extracted key-value pairs.
        """
        # Define the regex pattern to match {<key>:<value>}
        pattern = r"\{([^{}:]+):([^{}]+)\}"

        # Find all matches in the input string
        matches = re.findall(pattern, input_string)

        # Create a dictionary from the matches
        result = {key.strip(): value.strip() for key, value in matches}
        return result

    @staticmethod
    def replace_placeholders(input_string: str, replacements: dict[str, str]) -> str:
        """
        Replace placeholders in the input string with values from the replacements dictionary.

        :param input_string: String containing placeholders in the format {variable_key: value}.
        :param replacements: Dictionary containing replacement values for each variable_key.
        :return: String with placeholders replaced by corresponding values.
        """
        # Define the pattern to match the placeholders {variable_key: value} and ignore those without a colon
        pattern = r"\{([^{}:]+):([^{}]+)\}"

        # Function to replace each match with the corresponding value from the replacements dictionary
        def replacer(match):
            variable_key = match.group(1).strip()  # Extract and strip the variable_key
            # Get the replacement value from the dictionary or use the original value if key not found
            return replacements.get(variable_key, match.group(0))

        # Use re.sub with the replacer function to replace all valid placeholders
        return re.sub(pattern, replacer, input_string)

    def process_string(self, value: str, key: str) -> str:
        """
        Process a string by translating and formatting it.

        :param value: String to be processed.
        :param key: Key associated with the string in the JSON data.
        :return: Translated and formatted string.
        """
        # Translate the value
        translated_value = self.translate(
            value, target_language=self.user_or_account.language
        )
        # Format the string using the corresponding kwargs_map
        formatted_value = translated_value.format_map(self.kwargs_map.get(key, {}))
        return formatted_value

    def process_dict(self, value: dict[str, Any], key: str) -> str:
        """
        Process a dictionary containing plainText and meta keys.

        :param value: Dictionary to be processed.
        :param key: Key associated with the dictionary in the JSON data.
        :return: Processed string with translated and formatted content.
        """
        plain_text = value["plainText"]
        # Translate the value
        translated_value: str = self.translate(
            plain_text, target_language=self.user_or_account.language
        )
        meta = value.get("meta", {})
        final_text: str = self.process_meta(translated_value, meta, key)

        try:
            # Format the string using the corresponding kwargs_map
            formatted_value = final_text.format_map(self.kwargs_map.get(key, {}))
        except KeyError as err:
            raise KeyError(
                f"no data for {str(err)} in kwargs_map['{key}']: meta: {meta}, final_text: {final_text}"
            ) from err
        return formatted_value

    @staticmethod
    def custom_format_map(template: str, replacements: dict[str, str]) -> str:
        """
        Custom format function that replaces placeholders in the format
        [key] with values from the replacements dictionary.

        :param template: The template string containing placeholders in the format [key].
        :param replacements: A dictionary containing replacement values for each key.
        :return: The formatted string with placeholders replaced by corresponding values.
        """
        # Define the pattern to match placeholders in the format [key]
        pattern = r"\[([^\[\]]+)\]"

        # Function to replace each match with the corresponding value from the replacements dictionary
        def replacer(match):
            key = match.group(1)  # Extract the key from the match
            return replacements.get(
                key, match.group(0)
            )  # Return the replacement value or the original placeholder

        # Use re.sub with the replacer function to replace all placeholders
        return re.sub(pattern, replacer, template)

    def process_meta(
        self, plain_text: str, meta: dict[str, dict[str, dict[str, str]]], key: str
    ) -> str:
        """
        Process meta information to replace placeholders with HTML tags.

        :param plain_text: Translated plain text string.
        :param meta: Meta information containing HTML tag and attributes.
        :param key: Key associated with the meta data.
        :return: String with placeholders replaced by HTML tags.
        """
        mapping = {
            _key: self.custom_format_map(
                value, replacements=self.kwargs_map.get(key, {})
            )
            for _key, value in self.extract_patterns(plain_text).items()
        }
        html_mapping = dict()
        for tag_variable, attributes in meta.items():
            tag, variable = tag_variable.split(":")
            # Format each attribute in the meta dictionary
            for attr, val in attributes["attr"].items():
                try:
                    attributes["attr"][attr] = val.format_map(
                        self.kwargs_map.get(key, {})
                    )
                except KeyError as e:
                    raise KeyError(f"No {str(e)} in kwargs_map['{key}']") from e
            variable = variable.strip()
            attributes_str = " ".join(
                [f'{k}="{v}"' for k, v in attributes["attr"].items()]
            )
            if variable in mapping:
                html_mapping[variable] = (
                    f"<{tag} {attributes_str}>" + "{" + variable + "}" + f"</{tag}>"
                ).format_map({variable: mapping[variable]})
        return self.replace_placeholders(
            input_string=plain_text, replacements=html_mapping
        )

    def get_context(self, **kwargs) -> dict[str, Any]:
        """
        Transform the JSON data by processing strings and dictionaries.

        :return: Dictionary containing transformed JSON data.
        """
        # Process the json_data
        result = {}
        for key, value in self.data.items():
            if isinstance(value, str):
                # Process the string value
                result[key] = self.process_string(value, key)
            elif isinstance(value, dict):
                # Handle plainText and meta structure
                result[key] = self.process_dict(value, key)
            else:
                raise ValueError(f"Value must either be a string or a dict: {value}")
        return {**result, **kwargs}



from django.conf import settings
from rest_framework.routers import DefaultRouter, SimpleRouter

if settings.DEBUG:
    router = DefaultRouter()
else:
    router = SimpleRouter()


# router.register("accounts", views.AccountViewSet, basename="accounts")


app_name = "learngual_v1"
urlpatterns = router.urls



import importlib
import logging
import os
import urllib.parse
from collections import OrderedDict
from logging import getLogger
from typing import Any, Literal
from urllib.parse import urlparse

import jwt
import pytz
import requests
from django.conf import settings
from django.core.cache import cache
from django.utils import timezone
from faker import Faker

from .translator import Translator

faker = Faker()

logger = getLogger(__file__)

LEARNGUAL_SERVICE_API_KEY = getattr(
    settings, "LEARNGUAL_SERVICE_API_KEY", None
) or os.getenv("LEARNGUAL_SERVICE_API_KEY", None)


LOCAL_URLS = {
    "iam": "http://host.docker.internal:8000",
    "learn": "http://host.docker.internal:8001",
    "payment": "http://host.docker.internal:8002",
    "notify": "http://host.docker.internal:8003",
    "media": "http://host.docker.internal:8004",
}


def transform_local_base_url(
    base_url: str,
    service: Literal["payment", "iam", "learn", "notify", "media"] = "iam",
) -> str:
    """check if baser_url is local url return the approprate url

    Args:
        base_url (str): _description_
        service
              _description_. Defaults to "iam".

    Returns:
        str: _description_
    """
    if not base_url.startswith("https://"):
        return LOCAL_URLS.get(service, "iam")
    return base_url


def get_service_request_headers(**kwargs) -> dict:
    """function add headers needed for request made from a service

    Returns:
        dict: _description_
    """
    if LEARNGUAL_SERVICE_API_KEY:
        kwargs["service-key"] = LEARNGUAL_SERVICE_API_KEY

    return {**kwargs}


def get_service_request_params(**kwargs) -> str:
    """function return query params needed to make request as a service

    Returns:
        dict: _description_
    """

    if LEARNGUAL_SERVICE_API_KEY:
        kwargs["_service-key"] = LEARNGUAL_SERVICE_API_KEY
    if not kwargs:
        return ""
    return "?" + "&".join([f"{x}={y}" for x, y in kwargs.items()])


def get_nested_value(data: dict[str, Any], path: str):
    """
    Retrieve a nested dictionary value using a dot path, including support for accessing lists and slicing.

    Args:
        data (Dict[str, Any]): The nested dictionary to traverse.
        path (str): The dot-separated path to the desired value.

    Returns:
        The value at the specified path if found, otherwise None.

    Example:
        data = {
            'foo': {
                'bar': [
                    {'baz': 42},
                    {'qux': [1, 2, 3, 4, 5]}
                ]
            }
        }

        result = get_nested_value(data, 'foo.bar[0].baz')
        # Output: 42

        result = get_nested_value(data, 'foo.bar[1].qux[2]')
        # Output: 3

        result = get_nested_value(data, 'foo.bar[1].qux[1:4]')
        # Output: [2, 3, 4]

        result = get_nested_value(data, 'foo.bar[1].qux[:3]')
        # Output: [1, 2, 3]

        result = get_nested_value(data, 'foo.bar[1].qux[2:]')
        # Output: [3, 4, 5]

        result = get_nested_value(data, 'foo.bar[1].qux[5]')
        # Output: None (Index out of range)

        result = get_nested_value(data, 'foo.bar[1].qux[4:2]')
        # Output: None (Invalid slice range)
    """
    keys = path.split(".")
    value = data

    try:
        for key in keys:
            if key.endswith("]"):
                key, index_or_slice = key[:-1].split("[")
                if ":" in index_or_slice:
                    start, stop = map(int, index_or_slice.split(":"))
                    value = value[key][start:stop]
                else:
                    index = int(index_or_slice)
                    value = value[key][index]
            else:
                value = value[key]
    except (KeyError, TypeError, IndexError, ValueError):
        value = None
        logging.warning(f"Could not retrieve path:'{path}'.")

    return value


def update_nested_value(data: dict[str, Any], path: str, value: Any) -> dict[str, Any]:
    """
    Update a nested dictionary value using a dot path and return the modified dictionary.

    Args:
        data (Dict[str, Any]): The nested dictionary to update.
        path (str): The dot-separated path to the value to update.
        value (Any): The new value to assign.

    Returns:
        Dict[str, Any]: The modified dictionary.

    Example:
        data = {
            'foo': {
                'bar': {
                    'baz': 42
                }
            }
        }

        updated_data = update_nested_value(data, 'foo.bar.baz', 99)
        # Now, updated_data['foo']['bar']['baz'] is 99

        updated_data = update_nested_value(data, 'foo.bar.qux', [1, 2, 3])
        # Now, updated_data['foo']['bar']['qux'] is [1, 2, 3]
    """
    keys = path.split(".")
    current_dict = data

    for key in keys[:-1]:
        if key not in current_dict or not isinstance(current_dict[key], dict):
            current_dict[key] = {}
        current_dict = current_dict[key]

    current_dict[keys[-1]] = value
    return data


def extract_base_url(url):
    parsed_url = urlparse(url)
    base_url = parsed_url.scheme + "://" + parsed_url.netloc
    return base_url


def flatten_dict(
    data: dict[str, Any], parent_key: str = "", sep: str = "."
) -> dict[str, Any]:
    """
    Flatten a nested dictionary into a new dictionary with dot path keys.

    Args:
        data (Dict[str, Any]): The nested dictionary to flatten.
        parent_key (str): The parent key to use for the current level of the dictionary (used recursively).
        sep (str): The separator to use between the parent key and the current key.

    Returns:
        Dict[str, Any]: The flattened dictionary with dot path keys.

    Example:
        data = {
            'foo': {
                'bar': {
                    'baz': 42
                },
                'qux': [1, 2, 3]
            },
            'hello': 'world'
        }

        flattened_data = flatten_dict(data)
        # flattened_data is:
        # {
        #     'foo.bar.baz': 42,
        #     'foo.qux': [1, 2, 3],
        #     'hello': 'world'
        # }
    """
    flattened = {}
    for key, value in data.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            flattened.update(flatten_dict(value, new_key, sep=sep))
        else:
            flattened[new_key] = value
    return flattened


def unflatten_dict(data: dict[str, Any], sep: str = ".") -> dict[str, Any]:
    """
    Convert a dictionary with dot path keys to a nested dictionary.

    Args:
        data (Dict[str, Any]): The dictionary with dot path keys to convert.
        sep (str): The separator used in the dot path keys.

    Returns:
        Dict[str, Any]: The nested dictionary.

    Example:
        data = {
            'foo.bar.baz': 42,
            'foo.qux': [1, 2, 3],
            'hello': 'world'
        }

        nested_data = unflatten_dict(data)
        # nested_data is:
        # {
        #     'foo': {
        #         'bar': {
        #             'baz': 42
        #         },
        #         'qux': [1, 2, 3]
        #     },
        #     'hello': 'world'
        # }
    """
    nested = {}
    for key, value in data.items():
        parts = key.split(sep)
        current_dict = nested
        for part in parts[:-1]:
            if part not in current_dict:
                current_dict[part] = {}
            current_dict = current_dict[part]
        current_dict[parts[-1]] = value
    return nested


class PermissonUtils:
    def __init__(self, permission: dict) -> None:
        """
        Example: {
            "id":1223344
            "metadata":{

            }
        }

        Args:
            permission (dict): _description_
        """
        assert type(permission) == dict, "permssion must be a dictionary"
        self.__permission = permission

    def to_dict(self) -> dict:
        """return a dictionary of modified permission

        Returns:
            dict: _description_
        """
        return self.__permission

    def to_flat_dict(
        self, parent_key: str = "", sep: str = ".", *args, **kwargs
    ) -> dict:
        """return a flat dictionary of modified permission

        Example:
        data = {
            'foo': {
                'bar': {
                    'baz': 42
                },
                'qux': [1, 2, 3]
            },
            'hello': 'world'
        }

        flattened_data = flatten_dict(data)
        # flattened_data is:
        # {
        #     'foo.bar.baz': 42,
        #     'foo.qux': [1, 2, 3],
        #     'hello': 'world'
        # }

        Returns:
            dict: _description_
        """
        return flatten_dict(
            self.to_dict(), parent_key=parent_key, sep=sep, *args, **kwargs
        )

    def bool(self, path: str):
        """retrieve a boolean value from a nested dictionary

        Args:
            path (str): Example: metadata.manage_course.value

        Returns:
            bool|None: _description_
        """
        res = get_nested_value(self.__permission, path)
        if res is not None:
            return str(res).strip().lower() in ["true", "1"] or res

    def int(self, path: str) -> int:
        """retrieve a boolean value from a nested dictionary

        Args:
            path (str): Example: metadata.request_count.value

        Returns:
            int: _description_
        """
        res = get_nested_value(self.__permission, path)
        try:
            return int(res or 0)
        except (ValueError, TypeError):
            return int()

    def float(self, path: str) -> float:
        """retrieve a boolean value from a nested dictionary

        Args:
            path (str): Example: metadata.audio_seconds.value

        Returns:
            float: _description_
        """
        res = get_nested_value(self.__permission, path)
        try:
            return float(res or 0)
        except (ValueError, TypeError):
            return float()

    def set_value(self, path: str, value: Any, force_create: bool = False) -> dict:
        """function is used to overwrite key in the permission

        Args:
            path (str): _description_
            value Any: Example: 10, -30, {"age":12}
            force_create (bool, optional): _description_. Defaults to False.

        Raises:
            KeyError: if key does not exist and force_create is equal to false

        Returns:
            dict: _description_
        """

        res = get_nested_value(self.__permission, path)

        if res is None and not force_create:
            raise KeyError(f"{path} does not exists")
        return update_nested_value(self.__permission, path, value)

    def add_number(self, path: str, number, force_create: bool = False) -> dict:  # noqa
        """function is used to increment or decrement

        Args:
            path (str): _description_
            number (float | int): Example: 10, -30
            force_create (bool, optional): _description_. Defaults to False.

        Raises:
            TypeError: if wrong type is passed as number
            KeyError: if key does not exist and force_create is equal to false

        Returns:
            dict: _description_
        """
        if type(number) == str and str(number).isdigit():
            number = float(number)

        res = get_nested_value(self.__permission, path)

        if res is None and not force_create:
            raise KeyError(f"{path} does not exists")

        if type(number) in [int, float]:
            data = (res + number) if type(res) in [int, float] else 0 + number
            return update_nested_value(self.__permission, path, data)
        else:
            raise TypeError(f"{number} must be of type int ot float")


class PermissionManager:
    def update_permission_with_event(
        self, event_name, routing_key, permission_data, permission_id
    ):
        ...

    def update_permission_with_api(
        self,
        base_url: str,
        permission_id,
        service: Literal["iam", "payment", "notify", "learn", "media"] = "iam",
        permmission_data=dict(),
        headers: dict = dict(),
        params: str = "",
    ):
        """_summary_

        Args:
            base_url (str): _description_
            permission_id (_type_): _description_
            service (Literal["iam", "payment", "notify", "learn", "media"], optional): _description_. Defaults to "iam".
            permmission_data (_type_, optional): _description_. Defaults to dict().
            headers (dict, optional): _description_. Defaults to dict().
            params (str, optional): e.g name=aka&age=201. Defaults to "".

        Raises:
            requests.exceptions.RequestException: _description_

        Returns:
            _type_: _description_
        """
        if not str(base_url or "").startswith("https://"):
            base_url = LOCAL_URLS.get(service) or LOCAL_URLS.get("iam")

        params = params.strip("?")
        params = dict([x.split("=") for x in params.split("&")])
        headers = get_service_request_headers(**headers)

        url_path = f"/{service}/v1/permissions/{permission_id}/"
        res = requests.patch(
            base_url.rstrip("/") + url_path + get_service_request_params(**params),
            json=permmission_data,
            headers=headers,
        )
        if not res.ok:
            try:
                response_detail = res.json().get("detail")
            except requests.JSONDecodeError:
                response_detail = None

            if response_detail == "Invalid service key":
                logger.error(f"invalid service key, headers={headers}")
                raise requests.exceptions.RequestException("invalid service key")
            elif response_detail == "account does not exist":
                logger.error(
                    f"Account does not exist, {response_detail =}, {service =}"
                )
                raise requests.exceptions.RequestException(
                    f"Account does not exist in {service} service."
                )
            else:
                logger.error("permission service is down %s", res.content)
                raise requests.exceptions.RequestException(
                    dict(error="permission service is down", response=res.content)
                )
        return res.json()

    def clear_cache(
        self,
        *,
        service: Literal["iam", "payment", "notify", "learn", "media"] = "iam",
        permission_id: str,
    ):
        """clear cache

        Args:
            permission_id (str): _description_
            service (Literal["iam", "payment", "notify", "learn", "media"], optional): _description_. Defaults to "iam".
        """
        url_path = f"/{service}/v1/service/permissions/{permission_id}/"
        cache.delete(url_path)

    def retrieve_permission(
        self,
        *,
        base_url: str,
        permission_id: str,
        service: Literal["iam", "payment", "notify", "learn", "media"] = "iam",
        dot_path: str = None,
        headers: dict = dict(),
        params: str = "",
        cache_timeout: float = timezone.timedelta(seconds=10).total_seconds(),
    ):
        """_summary_

        Args:
            permission_id (str): example: '123456', 'sdGh66gGGHgfadsty', 'product:1234567'
            service (_type_): "iam" | "pay" | "notify" | "learn" | "media"
            base_url (str):
            dot_path (str):Default:None, e.g metadata.request_count.value
            headers (dict):Default:{}
            params (str):Default:"", e.g name=ann&age=102
            cache_timeout (float):Default:timezone.timedelta(hours=1).total_seconds()
        """
        url_path = f"/{service}/v1/service/permissions/{permission_id}/"
        if not str(base_url or "").startswith("https://"):
            base_url = LOCAL_URLS.get(service) or LOCAL_URLS.get("iam")
        data = cache.get(url_path)

        params = params.strip("?")
        params = dict([x.split("=") for x in params.split("&") if x])

        if not data:
            headers = get_service_request_headers(**headers)
            res = requests.get(
                base_url.rstrip("/") + url_path + get_service_request_params(**params),
                headers=headers,
            )
            if not res.ok:
                try:
                    response_detail = res.json().get("detail")
                except requests.JSONDecodeError:
                    response_detail = None

                if response_detail == "Invalid service key":
                    logger.error(f"invalid service key, headers={headers}")
                    raise requests.exceptions.RequestException("invalid service key")
                elif response_detail == "account does not exist":
                    logger.error(
                        f"Account does not exist, {response_detail =}, {service =}"
                    )
                    raise requests.exceptions.RequestException(
                        f"Account does not exist in {service} service."
                    )
                else:
                    logger.error("permission service is down %s", res.content)
                    raise requests.exceptions.RequestException(
                        dict(error="permission service is down", response=res.content)
                    )
            data = res.json()
            cache.set(url_path, data, timeout=cache_timeout)

        if dot_path:
            return get_nested_value(data, dot_path)
        return data


class TestHelper:
    def generate_timedelta(
        self,
        when: Literal["before", "after"],
        period: Literal["weeks", "days", "minutes", "seconds"] = "days",
        value: int = 2,
    ) -> str:
        """
        Args:
            when (Literal["before", "after"]): description
            period (Literal["weeks", "days", "minutes", "seconds"]): description
            value (int): description
        """
        if when == "before":
            return (
                (timezone.now() - timezone.timedelta(**{period: value}))
                .date()
                .isoformat()
            )
        elif when == "after":
            return (
                (timezone.now() + timezone.timedelta(**{period: value}))
                .date()
                .isoformat()
            )

    def no_duplicate(
        self, data: list[str | int] | list[dict[str, Any]], id_field: str | int = "id"
    ) -> bool:
        if not data:
            return True
        if type(data[0]) in [dict, OrderedDict]:
            data = [x.get(id_field) for x in data]
        return len(data) == len(set(data))

    def has_no_duplicate_in_response_results(
        self, response, id_field: str | int = "id"
    ) -> bool:
        data: list[str | int] | list[dict[str, Any]] = response.data.get("results")
        if not data:
            return True
        if type(data[0]) in [dict, OrderedDict]:
            data = [x.get(id_field) for x in data]
        return len(data) == len(set(data))

    def has_fields(self, data: dict, fields: list[int | str]) -> bool:
        conditions = []
        for x in fields:
            exist = x in data
            conditions.append(exist)
            if not exist:
                logging.warning("field -> '%s' does not exists", x)
        return all(conditions)

    def extract_results_in_response(self, response) -> list[dict]:
        return response.data.get("results")

    def has_fields_in_response_results(self, response, fields: list[int | str]) -> bool:
        results: list[dict] = response.data.get("results")
        if not results:
            return False
        data: dict = results[0]
        conditions = []
        for x in fields:
            exist = x in data
            conditions.append(exist)
            if not exist:
                logging.warning("field -> '%s' does not exists", x)
        return all(conditions)

    def has_paginated_count(self, response, count: int) -> bool:
        return response.data.get("count") == count

    def has_response_status(self, response, status_code: int) -> bool:
        return response.status_code == status_code

    def add_query_params_to_url(self, url: str, params: dict[str, Any]) -> str:
        query_string = urllib.parse.urlencode(params)
        return f"{url}?{query_string}"


def get_timezone_from_country(country_code) -> str:
    try:
        country_timezones = pytz.country_timezones.get(country_code.upper())
        if country_timezones:
            # Assuming the first timezone for simplicity
            return pytz.timezone(country_timezones[0])
        else:
            return None  # Country code not found or no timezone information available
    except KeyError:
        return None  # Invalid country code


def get_base_url(
    service: Literal["iam", "payment", "learn", "notify", "media"] = "iam"
):
    url: str = settings.LEARNGUAL_AUTH_RETRIEVE_URL
    if not url.startswith("https://"):
        url = LOCAL_URLS.get(service) or LOCAL_URLS.get("iam")

    parsed_url = urlparse(url)
    base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"
    return base_url


def extract_jwt_payload(refresh_token, secret_key):
    """Extracts the payload from a JWT refresh token.

    Args:
        refresh_token: The JWT refresh token.
        secret_key: The secret key used to decode the token.

    Returns:
        The decoded payload as a dictionary, or None if decoding fails.
    """

    try:
        decoded_token = jwt.decode(
            refresh_token, secret_key, algorithms=["HS256"]
        )  # Replace 'HS256' with your algorithm
        return decoded_token
    except jwt.exceptions.InvalidTokenError:
        return {}


def _translate(text: str, target_language: str = "EN", **kwargs):
    """Function to help translate micro-copies

    Args:
        text (str): _description_
        target_language (str): _description_

    Returns:
        _type_: _description_
    """
    return Translator().get_translation(text, target_language, **kwargs)


def extract_language_from_context(context: dict[str, Any]) -> str:
    """Function to extract language from serializer context

    Args:
        context (dict[str,Any]): _description_

    Returns:
        str: _description_
    """
    language = "EN"
    if context:
        request = context.get("request")
        if request:
            account = getattr(request, "account", None) or getattr(
                request.user, "account", None
            )
            if account:
                language = account.language
            elif request.user.is_authenticated:
                language = request.user.language
            else:
                language = request.GET.get("_lang")
    return language


def load_callable(path: str) -> object | None:
    paths = path.split(".")
    modules = importlib.import_module(".".join(paths[:-1]))
    result = getattr(modules, paths[-1], None)
    if not result:
        logger.warning("Module does no exists. path: %s", path)
    return result



from rest_framework import permissions

from . import custom_exceptions


class ExamplePermissions:
    class CanAccessView(permissions.BasePermission):
        def has_object_permission(self, request, view, obj):

            return False

        def has_permission(self, request, view):

            return True


class GeneralPermissions:
    class IsService(permissions.BasePermission):
        """if request is made from a service as a service

        Args:
            permissions (_type_): _description_
        """

        def has_object_permission(self, request, view, obj):

            return False

        def has_permission(self, request, view):
            return getattr(request.user, "is_service", False)

    class IsApiKey(permissions.BasePermission):
        """if request is made with an apikey

        Args:
            permissions (_type_): _description_
        """

        def has_permission(self, request, view):
            api_key = str(
                request.META.get("HTTP_API_KEY", "") or request.GET.get("api-key", "")
            )
            return bool(api_key)

    class IsNotApiKey(permissions.BasePermission):
        """if request is not made with an api key

        Args:
            permissions (_type_): _description_
        """

        def has_permission(self, request, view):
            api_key = str(
                request.META.get("HTTP_API_KEY", "") or request.GET.get("api-key", "")
            )
            return not bool(api_key) and request.user.is_authenticated


class IsAuthenticated(permissions.BasePermission):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self, request, view):
        if bool(request.user and request.user.is_authenticated):
            return True
        raise custom_exceptions.UnAuthenticated(
            "Authentication credentials were not provided"
        )



import importlib
import logging
import os
from logging import getLogger

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _
from rest_framework import authentication
from rest_framework.request import Request

from . import custom_exceptions

logger = getLogger(__file__)

User = get_user_model()

LEARNGUAL_SERVICE_API_KEY = getattr(
    settings, "LEARNGUAL_SERVICE_API_KEY", None
) or os.getenv("LEARNGUAL_SERVICE_API_KEY", None)

"""
Configure authentication class
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        ...
        "iam_service.users.authentication.LearngualAuthentication",
        ...
    ),
    ...
}

LEARNGUAL_AUTH_RETRIEVE_URL=<auth server url to authenticate if user is login>
LEARNGUAL_AUTH_TEST_MODE=<True or False used for testing provide dummy data to authenticate user>
LEARNGUAL_AUTH_GET_USER=<dot path to the callable >

Data = {
    "account": {
        "id": "84bcaf2972",
        "cover_photo": None,
        "profile_photo": None,
        "type": "PERSONNAL",
        "metadata": {},
        "created_at": "2023-01-13T16:33:52.084540Z",
        "updated_at": "2023-01-13T16:33:52.084576Z"
    },
    "email": "Bulah53@gmail.com",
    "first_name": "Caitlyn",
    "id": "40e0e7013f",
    "last_name": "Marquardt",
    "registration_step": "REGISTRATION_COMPLETED",
    "username": "Eloisa.Senger42"
}

def get_user(data:Data):

    print("data in get user")
    return get_user_model().objects.get_or_create()

"""


user_test_data = {
    "account": {
        "id": "84bcaf2972",
        "cover_photo": None,
        "profile_photo": None,
        "type": "PERSONAL",
        "metadata": {},
        "created_at": "2023-01-13T16:33:52.084540Z",
        "updated_at": "2023-01-13T16:33:52.084576Z",
    },
    "email": "Bulah53@gmail.com",
    "first_name": "Caitlyn",
    "id": "40e0e7013f",
    "last_name": "Marquardt",
    "registration_step": "REGISTRATION_COMPLETED",
    "username": "Eloisa.Senger42",
}


def load_callable(path: str) -> object | None:
    paths = path.split(".")
    modules = importlib.import_module(".".join(paths[:-1]))
    result = getattr(modules, paths[-1], None)
    if not result:
        logger.warning("Module does no exists. path: %s", path)
    return result


LEARNGUAL_AUTH_RETRIEVE_URL = getattr(settings, "LEARNGUAL_AUTH_RETRIEVE_URL", None)
LEARNGUAL_AUTH_GET_USER = getattr(settings, "LEARNGUAL_AUTH_GET_USER", None)
LEARNGUAL_AUTH_ACCOUNT_MODEL_PATH = getattr(
    settings, "LEARNGUAL_AUTH_ACCOUNT_MODEL_PATH", None
)


assert (
    LEARNGUAL_AUTH_RETRIEVE_URL
), "LEARNGUAL_AUTH_RETRIEVE_URL must be provided in the settings."
assert (
    LEARNGUAL_AUTH_GET_USER
), "LEARNGUAL_AUTH_GET_USER must be provided in the settings."


get_user: callable = load_callable(LEARNGUAL_AUTH_GET_USER)
assert get_user, "No callable exists in path:%s" % LEARNGUAL_AUTH_GET_USER


class LearngualAuthentication(authentication.BaseAuthentication):
    """
    An authentication plugin that authenticates requests through a JSON web
    token provided in a request header.
    """

    www_authenticate_realm = "api"
    media_type = "application/json"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_model = get_user_model()

    @classmethod
    def get_http_headers(cls, request: Request):
        headers = {
            key.lower().lstrip("http_"): value
            for key, value in request.META.items()
            if key.lower().startswith("http_")
        }
        return headers

    @classmethod
    def get_query_str(cls, request: Request):
        return "?{query}".format(query=request.META.get("QUERY_STRING", ""))

    def authenticate(self, request):
        # this is for test mode
        service_key = str(
            request.META.get("HTTP_SERVICE_KEY", "")
            or request.GET.get("service-key", "")
            or request.GET.get("_service-key", "")
        )
        if service_key:
            logger.info("authentication send with service account")
            if service_key == LEARNGUAL_SERVICE_API_KEY:
                service_user = User(id="service", username="service")
                service_user.is_service = True
                return service_user, service_key
            logger.warning("invalid service key")
            msg = _("Invalid service key")
            raise custom_exceptions.UnAuthenticated(msg)

        header = self.get_header(request)
        logger.info(
            "retrieve required authentication header during authentication -> %s",
            header,
        )
        if not header.get("authorization"):
            logger.warning("no authorization")
            return

        headers = self.get_http_headers(request)
        logger.info("get django request header for forwarding -> %s", headers)

        query_str = self.get_query_str(request)
        logger.info("query string for forwarding -> %s", query_str)
        logger.info(f"{LEARNGUAL_AUTH_RETRIEVE_URL =}")
        res = requests.get(LEARNGUAL_AUTH_RETRIEVE_URL + query_str, headers=headers)
        if not res.ok:
            logging.warning("request to IAM service did not go through")
            raise custom_exceptions.UnAuthenticated("Service is down.")
        res_data = res.json()
        return (
            get_user(res_data),
            header.get("api_key") or header.get("authorization").split(" ")[-1],
        )

    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        account_id = str(
            request.META.get("HTTP_ACCOUNT", "")
            or request.META.get("HTTP_X_ACCOUNT", "")
        )
        api_key = str(
            request.META.get("HTTP_API_KEY", "")
            or request.META.get("HTTP_X_API_KEY", "")
        )
        service_key = str(
            request.META.get("HTTP_SERVICE_KEY", "")
            or request.META.get("HTTP_X_SERVICE_KEY", "")
            or request.GET.get("service-key", "")
            or request.GET.get("_service-key", "")
        )
        authorization = str(request.META.get("HTTP_AUTHORIZATION", ""))

        return dict(
            account_id=account_id,
            api_key=api_key,
            authorization=authorization,
            service_key=service_key,
        )


def default_user_authentication_rule(user):
    # Prior to Django 1.10, inactive users could be authenticated with the
    # default `ModelBackend`.  As of Django 1.10, the `ModelBackend`
    # prevents inactive users from authenticating.  App designers can still
    # allow inactive users to authenticate by opting for the new
    # `AllowAllUsersModelBackend`.  However, we explicitly prevent inactive
    # users from authenticating to enforce a reasonable policy and provide
    # sensible backwards compatibility with older Django versions.
    return user is not None and user.is_active


class LearngualFakeAuthentication(authentication.BaseAuthentication):
    """
    A fake authentication plugin that authenticates requests through a JSON web
    token provided in a request header.
    It simulates authentication by the IAM service, but uses the "account"
    header to internally retrieve the account
    """

    www_authenticate_realm = "api"
    media_type = "application/json"
    account_model: type[models.Model] | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_model = get_user_model()
        self.account_model = load_callable(LEARNGUAL_AUTH_ACCOUNT_MODEL_PATH)

    @classmethod
    def get_http_headers(cls, request: Request):

        headers = dict()
        if request:
            for key, value in request.META.items():
                if key.lower().startswith("http_"):
                    headers[key.lower().lstrip("http_")] = value
        return headers

    @classmethod
    def get_query_str(cls, request: Request):
        return "?{query}".format(query=request.META.get("QUERY_STRING", ""))

    def authenticate(self, request):
        logger.info("authenticate() starting")
        header = self.get_header(request)

        headers = self.get_http_headers(request)

        account_or_user_id = headers.get("account")
        # logger.info("account_or_user_id: %s", account_or_user_id)
        account = None
        user = None
        if self.account_model:
            account = self.account_model.objects.filter(id=account_or_user_id).first()
            if account:
                logger.info("account exist: %s", account)
                user = getattr(account, "owner", None) or getattr(account, "user", None)
                user.account = account

        if not account:
            logger.info("account not found")
            user = self.user_model.objects.filter(id=account_or_user_id).first()
            if user:
                logger.info("user exists: %s", user)
                user.account = None

        if not user:
            return None
        logger.info("authenticate() completed")
        return (
            user,
            header.get("api_key")
            or (header.get("authorization").split(" ") or ["fake_token"])[-1],
        )

    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        account_id = str(request.META.get("HTTP_ACCOUNT", ""))
        api_key = str(request.META.get("HTTP_API_KEY", ""))
        authorization = str(request.META.get("HTTP_AUTHORIZATION", ""))

        return dict(account_id=account_id, api_key=api_key, authorization=authorization)



import importlib
import json
import os
from collections import defaultdict
from logging import getLogger

import gspread
from django.conf import settings
from django.core.cache import cache
from django.utils import timezone
from google.oauth2.service_account import Credentials
from rest_framework.serializers import ValidationError

from .enums import LanguageCodeType

logger = getLogger(__file__)

LEARNGUAL_CACHE_MODEL = settings.LEARNGUAL_CACHE_MODEL


def load_callable(path: str) -> object | None:
    paths = path.split(".")
    modules = importlib.import_module(".".join(paths[:-1]))
    result = getattr(modules, paths[-1], None)
    if not result:
        logger.warning("Module does no exists. path: %s", path)
    return result


"""
from iam_service.learngual.translator import Translator
t = self = Translator(target_language="KO")
t.log_queue
"""


class Translator:
    key = "Translator-datav2"
    queue_key = "Translator-queue"

    refresh_duration = timezone.timedelta(minutes=1).total_seconds()

    def __init__(
        self,
        sheet_id=None,
        sheet_name=None,
        celery_app_path: str | None = None,
        target_language: str = "EN",
    ):
        cache.delete_many(["Translator-data", "Translator-datav2", "Translator-queue"])
        self.sheet_id = (
            sheet_id
            or getattr(settings, "LEARNGUAL_TRANSLATE_SHEET_ID", None)
            or os.getenv("LEARNGUAL_TRANSLATE_SHEET_ID")
        )
        self.sheet_name = (
            sheet_name
            or getattr(settings, "LEARNGUAL_TRANSLATE_SHEET_NAME", None)
            or os.getenv("LEARNGUAL_TRANSLATE_SHEET_NAME")
        )

        self.celery_app_path = (
            celery_app_path
            or getattr(settings, "LEARNGUAL_CELERY_APP_PATH", None)
            or os.getenv("LEARNGUAL_CELERY_APP_PATH")
        )

        assert (
            self.sheet_id
        ), "`LEARNGUAL_TRANSLATE_SHEET_ID` must be set in enviroment variable or Django setting"
        assert (
            self.sheet_name
        ), "`LEARNGUAL_TRANSLATE_SHEET_NAME` must be set in enviroment variable or Django setting"

        scopes = ["https://www.googleapis.com/auth/spreadsheets"]
        creds = json.loads(
            getattr(settings, "LEARNGUAL_GOOGLE_BOT_GRED", None)
            or os.getenv("LEARNGUAL_GOOGLE_BOT_GRED")
            or "{}"
        )
        credentials = Credentials.from_service_account_info(creds, scopes=scopes)
        self.gc = gspread.authorize(credentials)

        self.target_language = target_language

    @property
    def translations(self):
        return self.get_cache_model().get_default_cache_model().translations

    @property
    def headers(self):
        return self.get_cache_model().get_default_cache_model().headers

    @property
    def celery_app(self):
        return load_callable(self.celery_app_path)

    def remove_duplicate_rows(self, column_name="en"):
        # Authenticate and open the Google Sheet

        sheet = self.gc.open_by_key(self.sheet_id)
        worksheet = sheet.worksheet(self.sheet_name)

        # Get all data from the sheet
        data = worksheet.get_all_values()

        # Extract header and data rows
        header = data[0]
        rows = data[1:]

        # Find the index of the target column
        try:
            if column_name.lower() in header:
                column_index = header.index(column_name.lower())
            elif column_name.upper() in header:
                column_index = header.index(column_name.upper())
            else:
                column_index = header.index(column_name)
        except ValueError:
            raise ValueError(f"Column '{column_name}' not found in the sheet.")

        # Use a dictionary to track the last occurrence of each value in the target column
        last_occurrence = defaultdict(int)

        # Iterate through the rows to find the last occurrence of each value
        for i, row in enumerate(rows):
            value = row[column_index]
            last_occurrence[value] = i + 1  # +1 because of header

        # Build a list of row indices to keep
        rows_to_keep = set(last_occurrence.values())

        # Delete rows that are not in the rows_to_keep set
        rows_to_delete = [
            i + 2 for i in range(len(rows)) if (i + 1) not in rows_to_keep
        ]
        rows_to_delete.reverse()  # Delete from bottom to top to avoid shifting issues

        for row_index in rows_to_delete:
            worksheet.delete_rows(row_index)

        logger.info(
            f"Removed {len(rows_to_delete)} duplicate rows from column '{column_name}'."
        )

    def get_cache_model(self):
        assert (
            LEARNGUAL_CACHE_MODEL
        ), "LEARNGUAL_CACHE_MODEL does not exists in the enviroment variable"
        return load_callable(LEARNGUAL_CACHE_MODEL)

    def update_translation(self, headers, translations):
        self.get_cache_model().get_default_cache_model().update_translations(
            translations=translations, headers=headers
        )

    def update_queue(self, *texts: str):
        return self.get_cache_model().get_default_cache_model().add_to_queue(*texts)

    def overwrite_queue(self, texts: list):
        model = self.get_cache_model().get_default_cache_model()
        model.queue = texts
        model.save()

    def force_load_translations(self):
        logger.info("Initiate refresh tranlsations")
        try:
            self.remove_duplicate_rows()
        except Exception:
            ...
        sheet = self.gc.open_by_key(self.sheet_id)
        worksheet = sheet.worksheet(self.sheet_name)
        data = worksheet.get_all_values()

        headers = [x.upper() for x in data[0]]
        translations = {}

        for row in data[1:]:
            en_value = row[0].strip()
            translations[en_value] = {
                headers[i]: row[i] for i in range(1, len(headers))
            }
        logger.info("Update Cache")
        self.update_translation(headers=headers, translations=translations)

        return translations

    def get_language_code(self, language: str) -> str:
        languages = {
            key.strip().upper(): value.strip().upper()
            for key, value in LanguageCodeType.dict_name_key().items()
        }
        if not language:
            language = "EN"
        language = language.strip().upper()
        return languages.get(language, language)

    @property
    def log_queue(self):
        return self.get_cache_model().get_default_cache_model().queue

    def log_microcopy(self, text: str):
        logger.info(f"Log microcopy: {text}")
        if text not in self.translations and text:
            translations = self.translations
            headers = self.headers
            translations[text] = {headers[i]: "" for i in range(1, len(headers))}
            self.update_translation(
                **{"headers": headers, "translations": translations}
            )

            if text not in self.log_queue:
                self.update_queue(text)
                logger.info(f"Add microcopy '{text}' to log queue.")

    def log_microcopy_job(self):
        queue_list = self.log_queue

        sheet = self.gc.open_by_key(self.sheet_id)
        worksheet = sheet.worksheet(self.sheet_name)
        data = worksheet.get_all_values()

        translations = {}
        headers = [x.upper() for x in data[0]]

        for row in data[1:]:
            en_value = row[0].strip()
            translations[en_value] = {
                headers[i]: row[i] for i in range(1, len(headers))
            }

        queue_list = [x for x in self.log_queue if x not in translations]
        batch_to_update = queue_list[:60]
        batch_to_keep = queue_list[61:]
        if self.log_queue:
            self.overwrite_queue(batch_to_keep)
        if batch_to_update:
            logger.info(f"Added {len(batch_to_update)} microcopies to queue to sheet")
            start_index = 2
            worksheet.insert_rows([[x] for x in batch_to_update], row=start_index)

        self.force_load_translations()

    def force_add_language_column(self, language: str):
        if self.headers and language not in self.headers:
            sheet = self.gc.open_by_key(self.sheet_id)
            worksheet = sheet.worksheet(self.sheet_name)
            # Add new column with target language header
            col_count = len(worksheet.row_values(1))
            worksheet.update_cell(1, col_count + 1, language.upper())
            logger.info(f"Added new language column: {language}")

    def add_language_column(self, language: str):
        if self.headers and language not in self.headers:
            headers: list = self.headers
            headers.append(language.upper())
            self.update_translation(
                **{"headers": headers, "translations": self.translations}
            )

            self.force_add_language_column(language=language)

    def translate(self, text: str, target_language: str):
        target_language = self.get_language_code(target_language)
        if target_language.upper() not in self.headers:
            if not self.headers:
                self.celery_app.send_task(
                    "iam.accounts.translate_load_translations",
                    routing_key="iam.accounts.translate_load_translations",
                )
            # else:
            #     self.add_language_column(target_language.upper())
            logger.info(f"Language: {target_language} does not exists in sheet")
            return text
        stripped_text = text.strip()
        translated_dict = self.translations.get(stripped_text)
        if not translated_dict:
            if translated_dict is None:
                self.log_microcopy(text)
            return stripped_text

        translated_text = translated_dict.get(target_language, text)
        if not translated_text:
            return stripped_text

        return translated_text

    def get_translation(
        self, text, target_language: str | None = None, **kwargs
    ) -> str:
        target_language = target_language or self.target_language
        assert target_language, "target_language is required."
        result = self.translate(text, target_language)
        if kwargs:
            result = self.render(result, **kwargs)

        return result

    def render(self, text: str, **kwargs):
        if not text:
            return text
        try:
            return text.format(**kwargs)
        except KeyError as e:
            raise ValidationError(
                {"detail": f"Imcomplete kwargs for {kwargs} for text {text}"}
            ) from e



from random import choice
from unittest.mock import patch
from uuid import uuid4

from faker import Faker

from ..utils import (
    PermissionManager,
    PermissonUtils,
    flatten_dict,
    get_nested_value,
    unflatten_dict,
    update_nested_value,
)
from .request_mock import requests

faker = Faker()


def test_update_nested_value():
    permission_data: dict = dict(
        metadata=dict(
            request_count=dict(
                value=faker.pyint(min_value=100, max_value=1000),
            ),
            manage_courses=dict(value=faker.pybool()),
            ids=[0, 1, 2, 3, 4, 5, 6],
            age=10,
        ),
    )
    res = update_nested_value(permission_data, "metadata.age", 20)
    assert res.get("metadata", {}).get("age") == 20
    assert permission_data.get("metadata", {}).get("age") == 20


def test_flatten_dict():
    permission_data: dict = dict(
        metadata=dict(
            request_count=dict(
                value=faker.pyint(min_value=100, max_value=1000),
            ),
            manage_courses=dict(value=faker.pybool()),
            ids=[0, 1, 2, 3, 4, 5, 6],
        ),
    )
    permission_data_flat = {
        "metadata.request_count.value": permission_data.get("metadata", {})
        .get("request_count", {})
        .get("value"),
        "metadata.manage_courses.value": permission_data.get("metadata", {})
        .get("manage_courses", {})
        .get("value"),
        "metadata.ids": permission_data.get("metadata", {}).get("ids"),
    }
    assert flatten_dict(permission_data) == permission_data_flat
    assert unflatten_dict(permission_data_flat) == permission_data


def test_get_nested_value():
    permission_data: dict = dict(
        metadata=dict(
            request_count=dict(
                value=faker.pyint(min_value=100, max_value=1000),
            ),
            manage_courses=dict(value=faker.pybool()),
            ids=[0, 1, 2, 3, 4, 5, 6],
        ),
    )
    assert get_nested_value(permission_data, "metadata.request_count.value")
    assert not get_nested_value(permission_data, "metadata.request_count.val")
    assert get_nested_value(permission_data, "metadata.request_count")
    assert (
        get_nested_value(permission_data, "metadata.ids[1]")
        == permission_data.get("metadata", {}).get("ids")[1]
    )
    assert (
        get_nested_value(permission_data, "metadata.ids[0:4]")
        == permission_data.get("metadata", {}).get("ids")[0:4]
    )


def test_PermissonUtils():
    permission_data: dict = dict(
        metadata=dict(
            request_count=dict(
                value=faker.pyint(min_value=100, max_value=1000),
            ),
            manage_courses=dict(value=faker.pybool()),
            value=20,
        ),
    )
    permission: PermissonUtils = PermissonUtils(permission_data)
    assert permission.bool("metadata.manage_courses.value") == permission_data.get(
        "metadata", {}
    ).get("manage_courses", {}).get("value")

    assert type(permission.bool("metadata.manage_courses.value")) == bool
    assert permission.bool("metadata.manage_courses.alue") is None

    assert permission.int("metadata.request_count.value") == permission_data.get(
        "metadata", {}
    ).get("request_count", {}).get("value")

    assert type(permission.int("metadata.request_count.value")) == int
    assert permission.int("metadata.request_count.alue") == int()

    assert type(permission.float("metadata.request_count.value")) == float

    # check increment works
    value = faker.pyint(min_value=1, max_value=100)
    res = permission.add_number("metadata.value", value)
    assert res.get("metadata", {}).get("value") == 20 + value

    # check that decrement works
    res = permission.add_number("metadata.value", -30)
    assert res.get("metadata", {}).get("value") == 20 + value - 30

    # check that error is raised if key does not exist
    try:
        permission.add_number("metadata.fake", -30)
        assert False, "did not raise key error"
    except KeyError:
        assert True

    # check that error is raised for wrong type
    try:
        permission.add_number("metadata.value", "334459aas")
        assert False, "did not raise type error"
    except TypeError:
        assert True

    # check that force create works
    assert permission.add_number("metadata.fake", -30, force_create=True)

    # check that force create works
    assert permission.set_value("metadata.fake2", -30, force_create=True)
    permission.set_value("metadata.value", -30)
    assert permission.to_dict().get("metadata", {}).get("value") == -30

    try:
        permission.set_value("metadata.fake4", -30)
        assert False
    except KeyError:
        assert True

    permission_data: dict = dict(
        metadata=dict(
            request_count=dict(
                value=faker.pyint(min_value=100, max_value=1000),
            ),
            manage_courses=dict(value=faker.pybool()),
            ids=[0, 1, 2, 3, 4, 5, 6],
        ),
    )
    permission_data_flat = {
        "metadata.request_count.value": permission_data.get("metadata", {})
        .get("request_count", {})
        .get("value"),
        "metadata.manage_courses.value": permission_data.get("metadata", {})
        .get("manage_courses", {})
        .get("value"),
        "metadata.ids": permission_data.get("metadata", {}).get("ids"),
    }
    assert PermissonUtils(permission_data).to_flat_dict() == permission_data_flat
    assert PermissonUtils(permission_data).to_dict() == permission_data


@patch("requests.get", return_value=requests.get)
def test_permission_manager(mock_requests):
    permission_manager: PermissionManager = PermissionManager()
    permission_flat_dict = {
        "metadata.request_count.value": 100000,
        "metadata.messages.value": True,
        "metadata.analytics.value": False,
    }
    permission_dict = unflatten_dict(permission_flat_dict)
    assert not permission_dict.get("metadata.request_count.value")
    mock_requests.return_value.json.return_value = permission_dict
    random_key = choice(list(permission_flat_dict.keys()))
    res = permission_manager.retrieve_permission(
        base_url="https://example.com", permission_id=uuid4().hex[:10], service="iam"
    )
    assert res == permission_dict, res
    res = permission_manager.retrieve_permission(
        base_url="https://example.com",
        permission_id=uuid4().hex[:10],
        service="iam",
        dot_path=random_key,
    )
    assert res == permission_flat_dict.get(random_key), res



from drf_spectacular.contrib.rest_framework_simplejwt import SimpleJWTScheme


class JWTScheme(SimpleJWTScheme):
    target_class = "learngual.authentication.LearngualAuthentication"
    name = ["JWTToken", "AccountId", "AccountIdX"]

    def get_security_definition(self, auto_schema):
        return [
            {
                "type": "http",
                "scheme": "bearer",
            },
            {"type": "apiKey", "in": "header", "name": "ACCOUNT"},
            {"type": "apiKey", "in": "header", "name": "X-ACCOUNT"},
        ]



from logging import getLogger
from typing import Any, Literal

import requests

from .utils import get_service_request_headers

logger = getLogger(__file__)


service_headers: dict[str, Any] = get_service_request_headers()

TYPE_SERVICES = Literal["learn", "iam", "payment", "media", "notify"]


def post(url, data=None, json=None, **kwargs):
    r"""Sends a POST request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs["headers"] = {
        **service_headers,
        **kwargs.get("headers", {}),
    }
    return requests.post(url, data=data, json=json, **kwargs)


def get(url, params=None, **kwargs):
    r"""Sends a GET request.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs["headers"] = {
        **service_headers,
        **kwargs.get("headers", {}),
    }
    return requests.get(url, params=params, **kwargs)


def put(url, data=None, **kwargs):
    r"""Sends a PUT request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs["headers"] = {
        **service_headers,
        **kwargs.get("headers", {}),
    }
    return requests.put(url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    r"""Sends a PATCH request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs["headers"] = {
        **service_headers,
        **kwargs.get("headers", {}),
    }
    return requests.patch(url, data=data, **kwargs)


def delete(url, **kwargs):
    r"""Sends a DELETE request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs["headers"] = {
        **service_headers,
        **kwargs.get("headers", {}),
    }
    return requests.delete(url, **kwargs)


class ServiceRequest:
    """
    Help to construct request with request headers for service request
    """

    def __init__(
        self, base_url: str, service: TYPE_SERVICES = "iam", version: str = "v1"
    ) -> None:
        self.service = service
        self.version = version
        self.base_url = base_url

    def construct_url(self, path: str) -> str:
        """generate service url with

        url = f"{base_url}/{service}/{version}/services/{path}/"

        Args:
            path (str): _description_

        Returns:
            str: _description_
        """
        base_url = self.base_url.rstrip("/")
        path = path.strip("/")
        url = f"{base_url}/{self.service}/{self.version}/services/{path}/"
        return url

    def post(self, path, data=None, json=None, **kwargs):
        r"""Sends a POST request.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json data to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :return: :class:`Response <Response>` object
        :rtype: requests.Response
        """
        return post(self.construct_url(path), data=data, json=json, **kwargs)

    def get(self, path, params=None, **kwargs):
        r"""Sends a GET request.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary, list of tuples or bytes to send
            in the query string for the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :return: :class:`Response <Response>` object
        :rtype: requests.Response
        """

        return get(self.construct_url(path), params=params, **kwargs)

    def put(self, path: str, data=None, **kwargs):
        r"""Sends a PUT request.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json data to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :return: :class:`Response <Response>` object
        :rtype: requests.Response
        """

        return put(self.construct_url(path), data=data, **kwargs)

    def patch(self, path, data=None, **kwargs):
        r"""Sends a PATCH request.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json data to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :return: :class:`Response <Response>` object
        :rtype: requests.Response
        """

        return patch(self.construct_url(path), data=data, **kwargs)

    def delete(self, path, **kwargs):
        r"""Sends a DELETE request.

        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :return: :class:`Response <Response>` object
        :rtype: requests.Response
        """

        return delete(self.construct_url(path), **kwargs)



import os
import shutil
from logging import getLogger
from pathlib import Path

import git
from django.conf import settings
from django.core.management.base import BaseCommand

logger = getLogger(__file__)


class Command(BaseCommand):
    help = "Download email template"

    # def add_arguments(self, parser):
    #     parser.add_argument('platform', nargs='+', type=str)

    def handle(self, *args, **options):
        token = getattr(
            settings, "LEARNGUAL_GITHUB_TOKEN", os.getenv("LEARNGUAL_GITHUB_TOKEN")
        )
        username = getattr(
            settings,
            "LEARNGUAL_GITHUB_USERNAME",
            os.getenv("LEARNGUAL_GITHUB_USERNAME"),
        )
        repo_name = getattr(
            settings,
            "LEARNGUAL_GITHUB_REPO_NAME",
            os.getenv("LEARNGUAL_GITHUB_REPO_NAME"),
        )
        branch_name = getattr(
            settings, "LEARNGUAL_GITHUB_BRANCH", os.getenv("LEARNGUAL_GITHUB_BRANCH")
        )
        assert all(
            [token, username, repo_name, branch_name]
        ), f'All {", ".join(["LEARNGUAL_GITHUB_TOKEN","LEARNGUAL_GITHUB_USERNAME","LEARNGUAL_GITHUB_REPO_NAME","LEARNGUAL_GITHUB_BRANCH"])} environment variable must be set.'  # noqa
        repo_url = f"https://{token}@github.com/{username}/{repo_name}.git"

        clone_to_path = settings.ROOT_DIR / "templates"

        self.clone_specific_branch(repo_url, branch_name, clone_to_path)
        self.stdout.write(self.style.SUCCESS("example command"))

    def clone_specific_branch(self, repo_url, branch_name, clone_to_path):
        try:
            clone_to_path = Path(clone_to_path)

            # Remove every file and folder in the clone_to_path before cloning
            if clone_to_path.exists() and clone_to_path.is_dir():
                shutil.rmtree(clone_to_path)
                self.stdout.write(
                    self.style.SUCCESS(
                        f"Cleared the contents of '{clone_to_path}' before cloning."
                    )
                )

            # Clone the specific branch
            git.Repo.clone_from(repo_url, clone_to_path, branch=branch_name)
            self.stdout.write(
                self.style.SUCCESS(
                    f"Branch '{branch_name}' has been cloned to '{clone_to_path}'."
                )
            )

            # Path to the .git directory using pathlib.Path
            git_dir = clone_to_path / ".git"

            # Remove the .git directory if it exists
            if git_dir.exists() and git_dir.is_dir():
                shutil.rmtree(git_dir)
                self.stdout.write(
                    self.style.SUCCESS(
                        f"The .git folder has been removed from '{clone_to_path}'."
                    )
                )

        except Exception as e:
            self.stdout.write(self.style.ERROR(f"An error occurred: {e}"))



from rest_framework import status
from rest_framework.exceptions import APIException


class ServiceUnavailable(APIException):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE
    default_detail = "Service temporarily unavailable, try again later."
    default_code = "service_unavailable"


class Forbidden(APIException):
    status_code = status.HTTP_403_FORBIDDEN
    default_detail = "forbidden"
    default_code = "forbidden"


class Badrequest(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = "bad request payload"
    default_code = "bad_request"


class NotAcceptable(APIException):
    status_code = status.HTTP_406_NOT_ACCEPTABLE
    default_detail = "not acceptable"
    default_code = "not_acceptable"


class MethodNotAllowed(APIException):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED
    default_detail = "not allow"
    default_code = "not_allowed"


class NotFound(APIException):
    status_code = status.HTTP_404_NOT_FOUND
    default_detail = "not found"
    default_code = "not_found"


class Redirect(APIException):
    status_code = status.HTTP_307_TEMPORARY_REDIRECT
    default_detail = "redirect"
    default_code = "redirect"


class UnAuthorized(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "unauthorized"
    default_code = "unauthorized"


class UnAuthenticated(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "unauthorized"
    default_code = "unauthorized"


class ProxyAuthenticationRequired(APIException):
    status_code = status.HTTP_407_PROXY_AUTHENTICATION_REQUIRED
    default_detail = "proxy authentication required"
    default_code = "proxy_authentication_required"



import zoneinfo
from logging import getLogger

import pytz
from django.utils import timezone

logger = getLogger(__file__)


class TimeZoneMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        tz_header = request.headers.get("TZ")

        if tz_header:
            try:
                timezone.activate(tz_header)
            except (pytz.UnknownTimeZoneError, zoneinfo.ZoneInfoNotFoundError):
                logger.error("Invalid timezone %s", tz_header)
                pass  # Handle unknown timezone error here
        else:
            # Set default timezone if TZ header is not provided
            timezone.activate("UTC")

        response = self.get_response(request)
        timezone.deactivate()
        return response



from unittest.mock import MagicMock

# Mocking the requests library
requests = MagicMock()

# Mocking the get() function
requests.get.return_value = MagicMock(status_code=200, text="Mocked response")
requests.get.return_value.json.return_value = {"key": "value"}
requests.get.return_value.ok = True

# Mocking the post() function
requests.post.return_value = MagicMock(status_code=201, text="Created")
requests.post.return_value.json.return_value = {"key": "value"}
requests.post.return_value.ok = True

# Mocking the put() function
requests.put.return_value = MagicMock(status_code=204, text="No Content")
requests.put.return_value.json.return_value = None
requests.put.return_value.ok = True

# Mocking the delete() function
requests.delete.return_value = MagicMock(status_code=204, text="No Content")
requests.delete.return_value.json.return_value = None
requests.delete.return_value.ok = True

# Mocking the patch() function
requests.patch.return_value = MagicMock(status_code=200, text="Updated")
requests.patch.return_value.json.return_value = {"key": "value"}
requests.patch.return_value.ok = True

# Mocking the head() function
requests.head.return_value = MagicMock(status_code=200, text="OK")
requests.head.return_value.json.return_value = None
requests.head.return_value.ok = True

# Mocking the options() function
requests.options.return_value = MagicMock(status_code=200, text="OK")
requests.options.return_value.json.return_value = None
requests.options.return_value.ok = True



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class LearngualConfig(AppConfig):
    name = "learngual"
    verbose_name = _("Learngual")

    def ready(self):
        try:
            import learngual.signals  # noqa F401
            import learngual.schema  # noqa F401
        except ImportError:
            pass


class IAMLearngualConfig(AppConfig):
    name = "iam_service.learngual"
    verbose_name = _("Learngual")

    def ready(self):
        try:
            import iam_service.learngual.signals  # noqa F401
            import iam_service.learngual.schema  # noqa F401
        except ImportError:
            pass



import urllib.parse
from uuid import uuid4

from django.db import models
from django.utils.translation import gettext_lazy as _


class ToDictChoiceHelper:
    @classmethod
    def dict(cls: type["models.TextChoices"]):
        return [
            {
                "id": uuid4().hex,
                "name": x[1],
                "value": x[0],
            }
            for x in cls.choices
        ]

    @classmethod
    def dict_name_key(cls: type["models.TextChoices"]):
        return {x[1]: x[0] for x in cls.choices}

    @classmethod
    def get_display_name(
        cls: type["models.TextChoices"], data: "LanguageCodeType"
    ) -> str:
        """Retrieve display name

        Args:
            data (LanguageCodeType): _description_

        Returns:
            str: _description_
        """
        choices = dict(cls.choices)
        return choices.get(data.value)


class LanguageCodeType(ToDictChoiceHelper, models.TextChoices):
    Azerbaijani = "az", _("Azerbaijani")
    Arabic = "ar", _("Arabic")
    Catalan = "ca", _("Catalan")
    Chinese_mandarin = "zh", _("Chinese (mandarin)")
    Czech = "cs", _("Czech")
    Danish = "da", _("Danish")
    Dutch = "nl", _("Dutch")
    English = "en", _("English")
    Esperanto = "eo", _("Esperanto")
    Finnish = "fi", _("Finnish")
    French = "fr", _("French")
    German = "de", _("German")
    Greek = "el", _("Greek")
    Hebrew = "he", _("Hebrew")
    Hindi = "hi", _("Hindi")
    Hungarian = "hu", _("Hungarian")
    Indonesian = "id", _("Indonesian")
    Irish = "ga", _("Irish")
    Italian = "it", _("Italian")
    Japanese = "ja", _("Japanese")
    Korean = "ko", _("Korean")
    Persian = "fa", _("Persian")
    Polish = "pl", _("Polish")
    Portuguese = "pt", _("Portuguese")
    Russian = "ru", _("Russian")
    Slovak = "sk", _("Slovak")
    Spanish = "es", _("Spanish")
    Swedish = "sv", _("Swedish")
    Turkish = "tr", _("Turkish")
    Ukrainian = "uk", _("Ukrainian")

    @classmethod
    def get_url(cls: type["LanguageCodeType"], data: "LanguageCodeType") -> str:
        """generate flag url

        Args:
            data (LanguageCodeType): _description_

        Returns:
            str: _description_
        """
        base_url = "https://learngual-bucket.sfo3.digitaloceanspaces.com/flags/%s.png"

        return base_url % urllib.parse.quote(
            cls.get_display_name(data) + "_" + data.value
        )



from django.db.models import Manager, QuerySet


class AccountManager:
    class AccountManager(Manager):
        def get_queryset(self) -> QuerySet:
            return AccountManager.AccountQuerySet(self.model, using=self._db).filter()

    class AccountQuerySet(QuerySet):
        def suggest(self, account, *args, **kwargs):
            return self.exclude(id=account.id).filter(*args, **kwargs)
